import ctypes

# ========== SDL2 Pathing ==========
import os
import sys
import platform

from pyjoystick.utils import PYJOYSTICK_DIR, change_path, rescale

# ========== SDL2 Pathing ==========
if 'windows' in platform.architecture()[1].lower():
    if '64' in platform.architecture()[0]:
        os.environ.setdefault('PYSDL2_DLL_PATH', os.path.join(PYJOYSTICK_DIR, 'sdl2_win64'))
    else:
        os.environ.setdefault('PYSDL2_DLL_PATH', os.path.join(PYJOYSTICK_DIR, 'sdl2_win32'))
# ========== END SDL2 Pathing ==========

with change_path(PYJOYSTICK_DIR):
    import sdl2

from pyjoystick.stash import Stash
from pyjoystick.interface import Key, Joystick as BaseJoystick


__all__ = ['Key', 'Joystick', 'run_event_loop', 'stop_event_wait',
           'get_init', 'init', 'quit', 'key_from_event',
           'get_guid', 'get_mapping', 'get_mapping_name', 'is_trigger']


class Joystick(BaseJoystick):
    @classmethod
    def get_joysticks(cls):
        # Check init
        if not get_init():
            init()

        return Stash(cls(i) for i in range(sdl2.SDL_NumJoysticks()))  # Use identifier not instance id.

    def __new__(cls, identifier=None, instance_id=None, *args, **kwargs):
        # Check init
        if not get_init():
            init()

        # Create the object
        joy = super().__new__(cls)

        if instance_id is not None:
            # Create the underlying joystick from the instance id.
            # SDL_JOYDEVICEREMOVED and all other SDL_JOY#### events give the instance id
            joy.joystick = sdl2.SDL_JoystickFromInstanceID(instance_id)
            # print('Instance ID:', raw_joystick, SDL_JoystickGetAttached(raw_joystick))
        else:
            # Create the underlying joystick from the enumerated identifier
            # SDL_JOYDEVICEADDED and SDL_NumJoysticks use open
            if identifier is None:
                identifier = 0
            if isinstance(identifier, str):
                # Get the joystick from the name or None if not found!
                for i in range(sdl2.SDL_NumJoysticks()):
                    raw_joystick = sdl2.SDL_JoystickOpen(i)
                    try:
                        if sdl2.SDL_JoystickName(raw_joystick).decode('utf-8') == identifier:
                            joy.joystick = raw_joystick
                            break
                    except:
                        pass
            else:
                joy.joystick = sdl2.SDL_JoystickOpen(identifier)
            # print('ID:', raw_joystick, SDL_JoystickGetAttached(raw_joystick))

        try:
            joy.identifier = sdl2.SDL_JoystickID(sdl2.SDL_JoystickInstanceID(joy.joystick)).value
            # joy.identifier = SDL_JoystickInstanceID(raw_joystick)
            joy.name = sdl2.SDL_JoystickName(joy.joystick).decode('utf-8')
            joy.numaxes = sdl2.SDL_JoystickNumAxes(joy.joystick)
            joy.numbuttons = sdl2.SDL_JoystickNumButtons(joy.joystick)
            joy.numhats = sdl2.SDL_JoystickNumHats(joy.joystick)
            joy.numballs = sdl2.SDL_JoystickNumBalls(joy.joystick)
            joy.init_keys()
        except:
            pass

        # Try to get the gamepad object
        try:
            joy.gamecontroller = sdl2.SDL_GameControllerOpen(joy.identifier)
            # FromInstanceId does not Attach!
            # joy.gamecontroller = SDL_GameControllerFromInstanceID(SDL_JoystickInstanceID(joy.joystick)
            # print('ID:', SDL_GameControllerGetAttached(joy.gamecontroller))
        except:
            joy.gamecontroller = None

        try:
            joy.guid = get_guid(joy.joystick)  # Using this is more reliable for the GameController stuff
        except:
            pass

        return joy

    def is_available(self):
        """Return if this joystick is still active and available."""
        try:
            return sdl2.SDL_JoystickGetAttached(self.joystick)
        except:
            return False

    def close(self):
        """Close the joystick."""
        try:
            sdl2.SDL_GameControllerClose(self.gamecontroller)
        except:
            pass
        try:
            sdl2.SDL_JoystickClose(self.joystick)
        except:
            pass


def get_init(module=sdl2.SDL_INIT_GAMECONTROLLER):
    """Return if the given module was initialized.

    Note:
        SDL_INIT_GAMECONTROLLER also initializes the joystick subsystem.
    """
    return sdl2.SDL_WasInit(module)


def init(module=sdl2.SDL_INIT_GAMECONTROLLER, *args, **kwargs):
    """Initialize the given module.

    Note:
        SDL_INIT_GAMECONTROLLER also initializes the joystick subsystem.
    """
    if get_init(module):
        quit(module)
    sdl2.SDL_Init(module)


def quit(module=sdl2.SDL_INIT_EVERYTHING):
    """Quit the given module.

    Note:
        SDL_INIT_GAMECONTROLLER also initializes the joystick subsystem.
    """
    try:
        sdl2.SDL_QUIT(module)
    except:
        pass


def get_guid(joystick):
    """Return the GUID from the given joystick object.

    Args:
        joystick (Joystick/SDL_Joystick): SDL2 joystick object

    Returns:
        guid (str): GUID String.
    """
    try:
        joystick = joystick.joystick
    except:
        pass
    guid = sdl2.SDL_JoystickGetGUID(joystick)
    guid_buff = ctypes.create_string_buffer(33)
    sdl2.SDL_JoystickGetGUIDString(guid, guid_buff, ctypes.sizeof(guid_buff))
    return guid_buff.value


def get_mapping(joystick):
    """Return the button mapping.

    Note:
        Hat keys have a value that is returned to make it easy to map a hat value to a function.

    Args:
        joystick (Joystick/str): Joystick object or String GUID

    Returns:
        d (dict): Dictionary of {name: Key} mappings
    """
    # Mapping dictionary
    mapping = {}
    map_str = None

    # ===== FROM GameController =====
    if map_str is None:
        try:
            map_str = sdl2.SDL_GameControllerMapping(joystick.gamecontroller)
        except:
            try:
                map_str = sdl2.SDL_GameControllerMapping(joystick)
            except:
                pass

    # ===== FROM GUID =====
    if map_str is None:
        try:
            guid = sdl2.SDL_JoystickGetGUIDFromString(joystick.guid)
            map_str = sdl2.SDL_GameControllerMappingForGUID(guid)
        except:
            try:
                guid = sdl2.SDL_JoystickGetGUID(joystick)
                map_str = sdl2.SDL_GameControllerMappingForGUID(guid)
            except:
                pass

    # Check the type
    if map_str is None:
        return mapping
    try:
        map_str = map_str.decode('utf-8')
    except:
        pass

    for item in map_str.split(','):
        if ":" in item:
            name, key = item.split(':', 1)
            if key.startswith('b'):
                mapping[name] = Key(Key.BUTTON, int(key[1:]), joystick=joystick)
            elif key.startswith('a'):
                mapping[name] = Key(Key.AXIS, int(key[1:]), joystick=joystick)
            elif key.startswith('h'):
                key, val = key.split('.', 1)
                mapping[name] = Key(Key.HAT, int(key[1:]), value=int(val), joystick=joystick)
    return mapping


def get_mapping_name(joystick, key):
    """Return the mapping name currently associated with this key."""
    for name, k in get_mapping(joystick).items():
        if key.keytype == k.keytype and key.number == k.number:
            if key.keytype == k.HAT:  # Hat also checks for the value
                if k.value == key.value:
                    return name
            else:
                return name
    return None


def is_trigger(joystick, axis_id):
    """Return if the given joystick axis is a trigger (Don't want triggers scaled from -1 resting to 1).

    Args:
          joystick (SDLJoystick): Joystick object
          axis_id (int): Axis number to check

    Returns:
          is_trigger (bool): If True this axis is a trigger.
    """
    try:
        if not sdl2.SDL_GameControllerGetAttached(joystick.gamecontroller):
            return 'trigger' in str(get_mapping_name(joystick, Key(Key.AXIS, axis_id))).lower()
    except:
        return False

    # Check the left trigger
    try:
        bind = sdl2.SDL_GameControllerGetBindForAxis(joystick.gamecontroller, sdl2.SDL_CONTROLLER_AXIS_TRIGGERLEFT)
        if bind.bindType == sdl2.SDL_CONTROLLER_BINDTYPE_AXIS and bind.value.axis == axis_id:
            return True
    except:
        try:
            return 'trigger' in str(get_mapping_name(joystick, Key(Key.AXIS, axis_id))).lower()
        except:
            pass

    # Check the right trigger
    try:
        bind = sdl2.SDL_GameControllerGetBindForAxis(joystick.gamecontroller, sdl2.SDL_CONTROLLER_AXIS_TRIGGERRIGHT)
        if bind.bindType == sdl2.SDL_CONTROLLER_BINDTYPE_AXIS and bind.value.axis == axis_id:
            return True
    except:
        try:
            return 'trigger' in str(get_mapping_name(joystick, Key(Key.AXIS, axis_id))).lower()
        except:
            pass

    return False


def key_from_event(event, joystick=None):
    """Every library type should implement a key_from_event function to convert an event into a key.

    Args:
        event (SDL_Event): Event that occurred
        joystick (Joystick)[None]: Joystick object

    Returns:
        key (Key)[None]: Key created from the event.
    """
    if joystick is None:
        try:
            joystick = Joystick(instance_id=event.jdevice.which)
        except (ValueError, TypeError, Exception):
            return None

    if event.type == sdl2.SDL_JOYBUTTONDOWN:
        return Key(Key.BUTTON, event.jbutton.button, 1, joystick)
    elif event.type == sdl2.SDL_JOYBUTTONUP:
        return Key(Key.BUTTON, event.jbutton.button, 0, joystick)
    elif event.type == sdl2.SDL_JOYAXISMOTION:
        value = event.jaxis.value
        if is_trigger(joystick, event.jaxis.axis):
            value = rescale(value, -32768, 32767, 0, 1)  # Make triggers rest at 0
        else:
            value = rescale(value, -32768, 32767, -1, 1)
        return Key(Key.AXIS, event.jaxis.axis, value, joystick)
    elif event.type == sdl2.SDL_JOYHATMOTION:
        return Key(Key.HAT, event.jhat.hat, event.jhat.value, joystick)
    elif event.type == sdl2.SDL_JOYBALLMOTION:
        # WIP (NOT TESTED)
        return Key(Key.BALL, event.jball.ball, event.jball.value, joystick)
    return None


def stop_event_wait():
    """Post an event to break out of the event loop wait."""
    try:
        user_event = sdl2.SDL_Event()
        user_event.type = sdl2.SDL_USEREVENT
        user_event.user.code = 2
        user_event.user.data1 = None
        user_event.user.data2 = None
        sdl2.SDL_PushEvent(ctypes.byref(user_event))
    except:
        pass


def run_event_loop(add_joystick, remove_joystick, handle_key_event, alive=None, **kwargs):
    """Run the an event loop to process SDL Events.

    Args:
        add_joystick (callable/function): Called when a new Joystick is found!
        remove_joystick (callable/function): Called when a Joystick is removed!
        handle_key_event (callable/function): Called when a new key event occurs!
        alive (callable/function)[None]: Function to return True to continue running. If None run forever
    """
    if alive is None:
        alive = lambda: True

    if not get_init():
        init()

    event = sdl2.SDL_Event()
    while alive():
        if sdl2.SDL_WaitEventTimeout(ctypes.byref(event), 2000) != 0:
            # Check the event
            if event.type == sdl2.SDL_JOYDEVICEADDED:
                try:
                    # NOTE: event.jdevice.which is the id to use for SDL_JoystickOpen()
                    joy = Joystick(identifier=event.jdevice.which)
                    add_joystick(joy)
                except:
                    pass
            elif event.type == sdl2.SDL_JOYDEVICEREMOVED:
                try:
                    # NOTE: event.jdevice.which is the id to use for SDL_JoystickFromInstanceID()
                    joy = Joystick(instance_id=event.jdevice.which)
                    remove_joystick(joy)
                except:
                    pass
            else:
                # NOTE: event.jdevice.which is the id to use for SDL_JoystickFromInstanceID()
                joy = Joystick(instance_id=event.jdevice.which)
                key = key_from_event(event, joy)
                if key is not None:
                    handle_key_event(key)


# Attach a way to stop waiting by posting an event
run_event_loop.stop_event_wait = stop_event_wait



"""
Thread Updater module to help update GUI items in a separate thread.
"""
from qt_thread_updater import get_updater, set_updater, cleanup_app, ThreadUpdater


__all__ = ['get_updater', 'set_updater', 'cleanup_app', 'ThreadUpdater']



import threading
from qtpy import QtWidgets, QtGui, QtCore

from pyjoystick.qt.widget_updater import get_updater


__all__ = ['AxisWidget', 'ButtonWidget', 'HatWidget', 'SpinSlider', 'LED']


class AxisWidget(QtWidgets.QWidget):
    def __init__(self, title=""):
        super().__init__()

        # Layout
        self.main_layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.main_layout)

        self._value = 0
        self.title = QtWidgets.QLabel(title)
        self.slider = SpinSlider(-100.0, 100.0, decimals=2)
        self.slider.setReadOnly(True)
        self.slider.setTickInterval(10)
        self.slider.setMinimumHeight(200)
        self.slider.setOrientation(QtCore.Qt.Vertical)
        self.slider.setValue(0)

        self.main_layout.addWidget(self.slider)
        self.main_layout.addWidget(self.title)
    # end Constructor

    def value(self):
        return self._value

    def setValue(self, value):
        self._value = value * 100
        get_updater().now_call_latest(self.slider.setValue, self._value)  # Only need latest value displayed


class ButtonWidget(QtWidgets.QWidget):
    def __init__(self, title=""):
        super().__init__()

        # Layout
        self.main_layout = QtWidgets.QHBoxLayout()
        self.setLayout(self.main_layout)

        self._value = 0
        self.title = QtWidgets.QLabel(title)
        self.led = LED()
        self.led.setValue(0)

        self.main_layout.addWidget(self.title)
        self.main_layout.addWidget(self.led)

        self.setSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum)

    def value(self):
        return self._value

    def setValue(self, value):
        self._value = value
        get_updater().now_call_in_main(self.led.setValue, self._value)  # Show every value change


class HatWidget(QtWidgets.QWidget):
    def __init__(self, title=""):
        super().__init__()

        # Layout
        self.main_layout = QtWidgets.QHBoxLayout()
        self.setLayout(self.main_layout)

        self._value = ""
        self.title = QtWidgets.QLabel(title)
        self.edit = QtWidgets.QLineEdit()
        self.edit.setMaximumWidth(100)
        self.edit.setReadOnly(True)

        self.main_layout.addWidget(self.title)
        self.main_layout.addWidget(self.edit)

        self.setSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum)

    def value(self):
        return self._value

    def setValue(self, value):
        self._value = str(value)
        get_updater().now_call_in_main(self.edit.setText, self._value)  # Show every value change


class SpinSlider(QtWidgets.QWidget):
    """Custom slider that displays the minimum, maximum, and spinbox next to the slider to aid in
    usability.

    Args:
        minimum (int)[None]: Minimum value
        maximum (int)[None]: Maximum value
        decimals (int)[0]: Number of decimals to display and control.
    """
    # Signals
    actionTriggered = QtCore.Signal(object)
    rangeChanged = QtCore.Signal(object, object)
    sliderMoved = QtCore.Signal(object)
    sliderPressed = QtCore.Signal()
    sliderReleased = QtCore.Signal()
    valueChanged = QtCore.Signal(object)

    # Slider class attributes
    TickPosition = QtWidgets.QSlider.TickPosition
    NoTicks = QtWidgets.QSlider.NoTicks
    TicksAbove = QtWidgets.QSlider.TicksAbove
    TicksBelow = QtWidgets.QSlider.TicksBelow
    TicksBothSides = QtWidgets.QSlider.TicksBothSides
    TicksLeft = QtWidgets.QSlider.TicksLeft
    TicksRight = QtWidgets.QSlider.TicksRight

    SliderAction = QtWidgets.QSlider.SliderAction
    SliderMove = QtWidgets.QSlider.SliderMove
    SliderNoAction = QtWidgets.QSlider.SliderNoAction
    SliderPageStepAdd = QtWidgets.QSlider.SliderPageStepAdd
    SliderPageStepSub = QtWidgets.QSlider.SliderPageStepSub
    SliderSingleStepAdd = QtWidgets.QSlider.SliderSingleStepAdd
    SliderSingleStepSub = QtWidgets.QSlider.SliderSingleStepSub
    SliderToMaximum = QtWidgets.QSlider.SliderToMaximum
    SliderToMinimum = QtWidgets.QSlider.SliderToMinimum

    SliderChange = QtWidgets.QSlider.SliderChange
    SliderOrientationChange = QtWidgets.QSlider.SliderOrientationChange
    SliderRangeChange = QtWidgets.QSlider.SliderRangeChange
    SliderStepsChange = QtWidgets.QSlider.SliderStepsChange
    SliderValueChange = QtWidgets.QSlider.SliderValueChange

    # SpinBox class attributes
    ButtonSymbols = QtWidgets.QDoubleSpinBox.ButtonSymbols
    NoButtons = QtWidgets.QDoubleSpinBox.NoButtons
    PlusMinus = QtWidgets.QDoubleSpinBox.PlusMinus
    UpDownArrows = QtWidgets.QDoubleSpinBox.UpDownArrows

    CorrectionMode = QtWidgets.QDoubleSpinBox.CorrectionMode
    CorrectToNearestValue = QtWidgets.QDoubleSpinBox.CorrectToNearestValue
    CorrectToPreviousValue = QtWidgets.QDoubleSpinBox.CorrectToPreviousValue

    StepEnabledFlag = QtWidgets.QDoubleSpinBox.StepEnabledFlag
    StepDownEnabled = QtWidgets.QDoubleSpinBox.StepDownEnabled
    StepNone = QtWidgets.QDoubleSpinBox.StepNone
    StepUPEnabled = QtWidgets.QDoubleSpinBox.StepUpEnabled

    def __init__(self, minimum=0, maximum=99, decimals=0, parent=None):
        super().__init__(parent)

        # Widgets
        self.spinbox = QtWidgets.QDoubleSpinBox()
        self.slider = QtWidgets.QSlider()
        self._min = QtWidgets.QLabel("0")
        self._max = QtWidgets.QLabel("99")
        self._value = 0

        # Check Inputs
        self.setDecimals(decimals)
        self.setRange(minimum, maximum)
        self.setValue(minimum)

        # Create the layout
        self.main_layout = QtWidgets.QBoxLayout(QtWidgets.QBoxLayout.LeftToRight)
        self.setLayout(self.main_layout)
        self.setContentsMargins(0, 0, 0, 0)
        self.main_layout.setContentsMargins(0, 0, 0, 0)

        self.main_layout.addWidget(self.spinbox)
        self.main_layout.addWidget(self._min)
        self.main_layout.addWidget(self.slider)
        self.main_layout.addWidget(self._max)

        # Style
        self.setOrientation(QtCore.Qt.Horizontal)
        self.slider.setFocusPolicy(QtCore.Qt.NoFocus)  # Keyboard up and down -> spinbox

        # Signals
        self.slider.actionTriggered.connect(self.actionTriggered.emit)
        self.slider.rangeChanged.connect(self.rangeChanged.emit)
        self.slider.sliderMoved.connect(self.sliderMoved.emit)
        self.slider.sliderPressed.connect(self.sliderPressed.emit)
        self.slider.sliderReleased.connect(self.sliderReleased.emit)
        self.slider.valueChanged.connect(self._value_changed)
        self.spinbox.valueChanged.connect(self._value_changed)

    def _value_changed(self, value):
        """Sync the spinbox and slider values."""
        if self.spinbox.value() != value:
            self.spinbox.blockSignals(True)
            self.spinbox.setValue(value)
            self.spinbox.blockSignals(False)
        if self.slider.value() != value:
            self.slider.blockSignals(True)
            self.slider.setValue(value)
            self.slider.blockSignals(False)
        self.valueChanged.emit(self.spinbox.value())

    def isReadOnly(self):
        return self.spinbox.isReadOnly()

    def setReadOnly(self, value):
        self.slider.setEnabled(not value)
        self.spinbox.setReadOnly(value)
        if value:
            self.spinbox.setButtonSymbols(QtWidgets.QDoubleSpinBox.NoButtons)
        else:
            self.spinbox.setButtonSymbols(QtWidgets.QDoubleSpinBox.UpDownArrows)

    def decimals(self):
        """Return the decimals.

        Returns:
            decimals (int): Number of decimals that are available. 0 Means a regular spin box is used.
        """
        return self.spinbox.decimals()

    def setDecimals(self, decimals):
        """Set the number of decimals.

        Args:
            decimals (int): Number of decimals to display.
        """
        self.spinbox.setDecimals(decimals)
        if decimals > 0:
            single_step = float("0."+str(1).zfill(decimals))
            self.spinbox.setSingleStep(single_step)
            self.slider.setSingleStep(single_step)
        elif self.spinbox.singleStep() < 1:
            self.spinbox.setSingleStep(1)
            self.slider.setSingleStep(1)

    def minimum(self):
        """Return the minimum range value."""
        return self.slider.minimum()

    def setMinimum(self, minimum):
        """Set the minimum range value.

         Note:
             The default value is 0.

        Args:
            minimum (int/flaot): Minimum value.
        """
        self.slider.setMinimum(minimum)
        self.spinbox.setMinimum(minimum)
        self._min.setText(str(self.minimum()))

    def maximum(self):
        """Return the maximum range value."""
        return self.slider.maximum()

    def setMaximum(self, maximum):
        """Set the maximum range value.

        Note:
             The default value is 99.

        Args:
            maximum (int/flaot): Maximum value.
        """
        self.slider.setMaximum(maximum)
        self.spinbox.setMaximum(maximum)
        self._max.setText(str(self.maximum()))

    def setRange(self, minimum, maximum):
        """Set the min and max range.

        Args:
            minimum (int/flaot): Minimum value.
            maximum (int/flaot): Maximum value.
        """
        # flip
        if minimum > maximum:
            temp = minimum
            minimum = maximum
            maximum = temp

        self.slider.setRange(minimum, maximum)
        self.spinbox.setRange(minimum, maximum)

        self._min.setText(str(self.minimum()))
        self._max.setText(str(self.maximum()))

    def singleStep(self):
        return self.spinbox.singleStep()

    def setSingleStep(self, value):
        self.spinbox.setSingleStep(value)
        self.slider.setSingleStep(value)

    def hasTracking(self):
        return self.slider.hasTracking()

    def setTracking(self, value):
        """Set if the valueChanged signal should be activated whenever the slider is moved."""
        self.slider.setTracking(value)
        self.spinbox.setKeyboardTracking(value)

    def specialValueText(self):
        """This property holds the special-value text.

        If set, the spin box will display this text instead of a numeric value whenever the
        current value is equal to minimum(). Typical use is to indicate that this choice has
        a special (default) meaning.

        See Also:
            QAbstractSpinBox.specialValueText
        """
        return self.spinbox.specialValueText()

    def setSpecialValueText(self, text):
        self.spinbox.setSpecialValueText(text)

    # ========== Slider Override ==========
    def tickInterval(self):
        return self.slider.tickInterval()

    def setTickInterval(self, ti):
        if self.tickPosition() == QtWidgets.QSlider.NoTicks:
            self.slider.setTickPosition(QtWidgets.QSlider.TicksAbove)
        self.slider.setTickInterval(ti)

    def tickPosition(self):
        return self.slider.tickPosition()

    def setTickPosition(self, position):
        self.slider.setTickPosition(position)

    def invertedAppearance(self):
        """Return if the appearance is inverted."""
        return self.slider.invertedAppearance()

    def setInvertedAppearance(self, value):
        """Set the slider inverted appearance.

        Args:
            value (bool): Is the appearance backwards?
        """
        self.slider.setInvertedAppearance(value)
        self.setOrientation(self.orientation())

    def invertedControls(self):
        return self.slider.invertedControls()

    def setInvertedControls(self, value):
        """Set if the keyboard controls for the slider should be inverted."""
        self.slider.setInvertedControls(value)

    def isSliderDown(self):
        """Return if the slider is pressed down."""
        return self.slider.isSliderDown()

    def setSliderDown(self, value):
        self.slider.setSliderDown(value)

    def orientation(self):
        """Return the slider orientation."""
        return self.slider.orientation()

    def setOrientation(self, orientation):
        """Set the orientation."""
        self.slider.setOrientation(orientation)
        if orientation == QtCore.Qt.Horizontal:
            self.main_layout.insertWidget(1, self._min)
            self.main_layout.insertWidget(3, self._max)
            if self.invertedAppearance():
                self.main_layout.setDirection(QtWidgets.QBoxLayout.RightToLeft)
            else:
                self.main_layout.setDirection(QtWidgets.QBoxLayout.LeftToRight)
        else:
            self.main_layout.insertWidget(3, self._min)
            self.main_layout.insertWidget(1, self._max)
            if self.invertedAppearance():
                self.main_layout.setDirection(QtWidgets.QBoxLayout.BottomToTop)
            else:
                self.main_layout.setDirection(QtWidgets.QBoxLayout.TopToBottom)

    def pageStep(self):
        return self.slider.pageStep()

    def setPageStep(self, value):
        self.slider.setPageStep(value)

    def repeatAction(self):
        return self.slider.repeatAction()

    def setRepeatAction(self, action, thresholdTime=5000, repeatTime=50):
        self.slider.setRepeatAction(action, thresholdTime, repeatTime)

    def sliderPosition(self):
        return self.slider.sliderPosition()

    def setSliderPosition(self, value):
        self.slider.setSliderPosition(value)

    def triggerAction(self, action):
        self.slider.triggerAction(action)

    def value(self):
        """Return the value."""
        return self._value

    def setValue(self, value):
        self._value = value
        get_updater().now_call_latest(self.slider.setValue, self._value)   # Only need latest value displayed
        get_updater().now_call_latest(self.spinbox.setValue, self._value)  # Only need latest value displayed

    def sliderChange(self, change):
        self.slider.sliderChange(change)

    # ========== SpinBox Override ==========
    def setCleanText(self, value):
        try:
            value = float(value)
        except: pass
        self.setValue(value)

    def cleanText(self):
        return self.spinbox.cleanText()

    def prefix(self):
        return self.spinbox.prefix()

    def setPrefix(self, prefix):
        self.spinbox.setPrefix(prefix)

    def suffix(self):
        return self.spinbox.suffix()

    def setSuffix(self, suffix):
        self.spinbox.setSuffix(suffix)

    def textFromValue(self, val):
        return self.spinbox.textFromValue(val)

    def valueFromText(self, text):
        return self.spinbox.valueFromText(text)


class LED(QtWidgets.QWidget):
    """Display for showing an LED light color.

    This class can be used to show a Text Label and LED light or just an LED light. This class is
    also capable of making the LED a clickable button.
    """

    clicked = QtCore.Signal()

    def __init__(self, state=None):
        super().__init__()
        self.setObjectName("LED")
        self.setMinimumSize(14, 14)

        # Properties
        self._state = None
        self._colors = {}
        self._value = QtGui.QColor(255, 0, 0)
        self.alert_time = 1000
        self._timer_arg = "blank"
        self._alert_timer = QtCore.QTimer()
        self._alert_timer.setSingleShot(True)
        self._alert_timer.timeout.connect(self.alert_timeout)

        # Button Properties
        self._seq_iter = None
        self._btn_seq = None

        # Add Default Colors
        self.addColor("red", QtGui.QColor(255, 0, 0))
        self.addColor(0, QtGui.QColor(255, 0, 0))
        self.addColor("off", QtGui.QColor(255, 0, 0))
        self.addColor("no", QtGui.QColor(255, 0, 0))

        self.addColor("green", QtGui.QColor(0, 255, 0))
        self.addColor(1, QtGui.QColor(0, 255, 0))
        self.addColor("on", QtGui.QColor(0, 255, 0))
        self.addColor("yes", QtGui.QColor(0, 255, 0))

        self.addColor("yellow", QtGui.QColor(255, 255, 0))
        self.addColor(2, QtGui.QColor(255, 255, 0))

        self.addColor("blank", QtGui.QColor(225, 225, 225))
        self.addColor("gray", QtGui.QColor(225, 225, 225))
        self.addColor("grey", QtGui.QColor(225, 225, 225))
        self.addColor("colorless", QtGui.QColor(225, 225, 225))
        self.addColor(3, QtGui.QColor(225, 225, 225))

        self.addColor("blue", QtGui.QColor(65, 65, 255))
        self.addColor(4, QtGui.QColor(65, 65, 255))

        self.addColor("orange", QtGui.QColor(255, 153, 0))
        self.addColor(5, QtGui.QColor(255, 153, 0))

        # self.setButtonSequence([0, 1])
        self.setState(state)

    def state(self):
        """Return the current state identifier of the LED."""
        return self._state

    def activeColor(self):
        """Return the active QColor of the LED."""
        return self._value

    def buttonSequence(self):
        """Return the button sequence list of states to change to on button click."""
        return self._btn_seq

    def value(self):
        """Gets the state. This method was made to be an interchangeable
        method with TextIndicator.
        """
        return self.state()

    def setValue(self, value):
        """Sets the LED State. This method was made to be an interchangeable
        method with TextIndicator.

        Args:
            value(key): Color key state to set the LED color.
        """
        self.setState(value)

    def setColor(self, value):
        self.setState(value)

    def setColors(self, colors):
        """A dictionary or list of colors.

        Args:
            colors(list):list of strings (rgb, rgba, or hex) or a list of QtGui.QColor's.
        """
        for color in colors:
            color = self.checkColor(color)
        self._colors = colors

        self.setState(self._state)

    def addColor(self, state, color):
        """Add a color to the color changing dictionary."""
        color = self.checkColor(color)
        self._colors[state] = color

    def setState(self, state):
        """Set the state of the LED and show the color for that state."""
        self._state = state
        try:
            color = self._colors[state]
        except KeyError:
            try:
                state = int(state)
            except(ValueError, TypeError):
                state = str(state)
            try:
                color = self._colors[state]
            except KeyError:
                color = self._colors["blank"]

        self._value = color
        self.update()

    @staticmethod
    def checkColor(color):
        """Check the given color and return a valid color."""
        valid_color = None
        if isinstance(color, str) and "," in color:
            valid_color = QtGui.QColor(color)

        elif isinstance(color, list):
            if len(color) == 4:
                valid_color = QtGui.QColor(color[0], color[1], color[2], color[3])
            else:
                valid_color = QtGui.QColor(color[0], color[1], color[2])

        elif isinstance(color, QtGui.QColor):
            valid_color = color

        return valid_color

    def setButtonSequence(self, sequence):
        """States for the color of the sequence. Allows button presses to cycle through a sequence.

        Args:
            sequence(iterable): Sequence of states when pressing the button.
        """
        self._btn_seq = sequence
        self._seq_iter = iter(self._btn_seq)

        self.clicked.connect(self.cycle_colors)

    def alert(self, timeout=None, color="red", timeout_color="blank"):
        """Change the color for a given time then change back to the timeout_color.

        Args:
            timeout (int)[None]: Time in seconds. If None use alert_time property
            color (str)["red"]: Color to immediately change to.
            timeout_color (str)["blank"]: Color to reset the LED to after the timeout.
        """
        if timeout is None:
            timeout = self.alert_time
        self.setColor(color)
        self._timer_arg = timeout_color

        self._alert_timer.stop()
        self._alert_timer.start(timeout)

    def alert_timeout(self):
        """Method to run after the alert time to clear the alert LED."""
        self.setState(self._timer_arg)

    def cycle_colors(self):
        """Cycle Colors."""
        try:
            next_state = next(self._seq_iter)
        except StopIteration:
            self._seq_iter = iter(self._btn_seq)
            next_state = next(self._seq_iter)
        # end

        self.setState(next_state)

    def mousePressEvent(self, event):
        """Override to activate button click and change to the next state."""
        if event.button() == QtCore.Qt.LeftButton:
            self.clicked.emit()

    def paintEvent(self, event):
        """Overrides the default paintEvent method to create the widget's
        display.

        Args:
            event: needed for paintEvent super class method override
        """
        painter = QtGui.QPainter(self)
        painter.setRenderHint(painter.Antialiasing, True)

        # Get Black Background
        rect_back = self.getBackgroundRect()
        center = rect_back.center()
        radius = rect_back.width() + 0.0

        # Draw Black Background
        grad = QtGui.QRadialGradient(center, radius)
        grad.setColorAt(0.0, QtGui.QColor(255, 255, 255, 10))
        grad.setColorAt(0.90, QtGui.QColor(0, 0, 0, 255))
        grad.setColorAt(0.98, QtGui.QColor(0, 0, 0, 100))
        grad.setColorAt(1.0, QtGui.QColor(0, 0, 0, 255))
        painter.setPen(QtGui.QColor(0, 0, 0))
        painter.setBrush(grad)
        painter.drawEllipse(center, radius, radius)

        # Get LED
        rect_val = self.getValueRect()
        center = rect_val.center()
        radius = rect_val.width() + 0.0

        # Draw LED
        red = self._value.red()
        green = self._value.green()
        blue = self._value.blue()
        grad = QtGui.QRadialGradient(center, radius)
        grad.setColorAt(0.0, QtGui.QColor(red, green, blue, 100))
        grad.setColorAt(0.95, QtGui.QColor(red, green, blue, 255))
        grad.setColorAt(1.0, QtGui.QColor(red, green, blue, 100))
        painter.setPen(self._value)
        painter.setBrush(grad)
        painter.drawEllipse(center, radius, radius)

        painter.end()

    def getValueRect(self):
        """Return the value rectangel container."""
        width = self.width()
        height = self.height()

        radius = min(width/2, height/2)
        center = QtCore.QPoint(width/2, height/2)
        back_r = radius - radius/50
        led_r = back_r - back_r/10
        rect_val = self.rect()

        # Set the LED
        rect_val.setWidth(led_r)
        rect_val.setHeight(led_r)
        rect_val.moveCenter(center)

        return rect_val

    def getBackgroundRect(self):
        """Return the background rectangel container."""
        width = self.width()
        height = self.height()

        radius = min(width/2, height/2)
        center = QtCore.QPoint(width/2, height/2)
        back_r = radius - radius/50
        rect_back = self.rect()

        # Set the LED
        rect_back.setWidth(back_r)
        rect_back.setHeight(back_r)
        rect_back.moveCenter(center)

        return rect_back



import os
import sys
import platform
import time
import threading
import contextlib


__all__ = ['is_py27', 'PYJOYSTICK_DIR', 'deadband', 'change_path', 'rescale', 'PeriodicThread']


is_py27 = sys.version_info < (3, 0)

PYJOYSTICK_DIR = os.path.dirname(sys.executable)
try:
    if getattr(sys, 'frozen', False):
        PYJOYSTICK_DIR = getattr(sys, '_MEIPASS', os.path.dirname(sys.executable))
    else:
        PYJOYSTICK_DIR = os.path.dirname(os.path.abspath(__file__))
except (NameError, AttributeError, Exception):
    pass


def deadband(val, dead=0.2, scale=1):
    """Return the deadband value for the controller Axis.

    Args:
        val (float): Raw controller value -1 to 1.
        dead (float)[0.2]: Deadband value
        scale (int) [1]: 100, 10, 1 indicates the range -100 to 100 ...
    """
    val = float(val)
    dead = abs(float(dead))

    # Check if the deadband is the same as the scale
    if dead == scale:
        dead = scale * 0.99

    if val >= dead:
        val -= dead
    elif val <= -dead:
        val += dead
    else:
        val = 0

    val = (val/(scale-dead)) * scale
    return val


@contextlib.contextmanager
def change_path(path):
    """Temporarily change the sys.path for imports"""
    try:
        ch_path = sys.path.index(path)
    except:
        ch_path = -1
        
    if ch_path != -1:
        sys.path.pop(ch_path)
    
    yield
    
    if ch_path != -1:
        sys.path.insert(ch_path, path)


def rescale(value, curr_min, curr_max, new_min, new_max):
    """Convert the value from one scale to a new scale.
    
    Args:
        value (int/float/object): Value to convert to the new scale
        curr_max (int/float): Current maximum value for the current scale.
        curr_min (int/float): Current minimum value for the current scale.
        new_max (int/float): New maximum value for the new scale.
        new_min (int/float): New minimum value for the new scale.
        
    Returns:
        value (int/float/object): New value that was converted to the new scale
    """
    return ((value - curr_min) / (curr_max - curr_min)) * (new_max - new_min) + new_min


class PeriodicThread(threading.Thread):
    def __init__(self, interval, target=None, name=None, args=None, kwargs=None, daemon=None):
        """Create a thread that will run a function periodically.
        Args:
            interval (int/float): How often to run a function in seconds.
        """
        self.interval = interval
        self.alive = threading.Event()
        if args is None:
            args = tuple()
        if kwargs is None:
            kwargs = dict()
        super(PeriodicThread, self).__init__(target=target, name=name, args=args, kwargs=kwargs, daemon=daemon)

        if is_py27:
            self._args = self.__args
            self._kwargs = self.__kwargs
            self._started = self.__started
            self._target = self.__target

        if self._target is None and hasattr(self, '_run'):
            if is_py27:
                self.__target = self._run
            self._target = self._run

    def start(self):
        """Start running the thread."""
        self.alive.set()
        if not self._started.is_set():
            super(PeriodicThread, self).start()

    def stop(self):
        """Stop running the thread."""
        try:
            self.alive.clear()
        except:
            pass

    def run(self):
        """The thread will loop through running the set _target method (default _run()). This
        method can be paused and restarted.
        """
        while self.alive.is_set():
            # Run the thread method
            start = time.time()
            self._target(*self._args, **self._kwargs)
            try:
                sleep = self.interval - (time.time() - start)
                if sleep > 0:
                    time.sleep(sleep)
            except ValueError:
                pass  # sleep time less than 0

    def join(self, timeout=None):
        """Join the thread closing it."""
        self.stop()
        super(PeriodicThread, self).join(timeout=timeout)

    def __enter__(self):
        """Enter statement for use of 'with' statement."""
        self.start()
        return self

    def __exit__(self, ttype, value, traceback):
        """Exit statement for use of the 'with' statement."""
        try:
            self.join(0)  # Make sure join has a 0 timeout so it is not blocking while exiting
        except RuntimeError:
            pass

        return ttype is None  # Return False if there was an error



"""
    slatcontroller.keys_trigger
    SeaLandAire Technologies
    @author: jengel

The BindKeys in this module work with two main objects an OrderedDict of keys called "keys". The
"keys" variable holds the key values and uses those values in the keys_pressed signal. The
"key_translator" object is a regular dictionary that helps take in user input keys and modify the
"keys" dictionary values.
"""
import os
import collections
import copy
import threading

from qtpy import QtCore, QtGui, QtWidgets

from pyjoystick.utils import PYJOYSTICK_DIR


__all__ = ["RESOURCE_DIR", "KEYBOARD_ACTIVE_IMG", "KEYBOARD_INACTIVE_IMG",
           "BindKeys", "WASDController", "ArrowController", "GimbalController"]


# ========== ICONS ==========
RESOURCE_DIR = os.path.join(PYJOYSTICK_DIR, 'resources')
KEYBOARD_ACTIVE_IMG = os.path.join(RESOURCE_DIR, "keyboard_active.png")
KEYBOARD_INACTIVE_IMG = os.path.join(RESOURCE_DIR, "keyboard_inactive.png")
# ========== END ICONS ==========


class BindKeys(QtCore.QObject):
    """Event Filter for binding keys.

    Example:
        DEFAULT_KEYS = collections.OrderedDict([("up", 0), ("down", 0), ("zoom", 0)])
        DEFAULT_TRANSLATOR = {QtCore.Qt.Key_Up: "up", QtCore.Qt.Key_Down: "down",
                              QtCore.Qt.Key_Minus: ("zoom", -1, True), QtCore.Qt.Key_Equal: ("zoom", 1, True)}

    Args:
        default_keys (OrderedDict)[None]: OrderedDictionary of ("key name", default value)
        key_translator (dict/NormalKey)[None]: Key translator
            {Actual key:  ("key name", modify_value, send_on_release, send_repeat)}
    """

    keys_pressed = QtCore.Signal(collections.OrderedDict)

    KEYBOARD_ACTIVE_IMG = KEYBOARD_ACTIVE_IMG
    KEYBOARD_INACTIVE_IMG = KEYBOARD_INACTIVE_IMG

    DEFAULT_KEYS = collections.OrderedDict()
    DEFAULT_TRANSLATOR = {} # Translate keyboard keys to key names or (key names, modifier)

    def __init__(self, key_values=None, key_translator=None):
        super().__init__()

        self._is_active = False
        self.key_translator = {}
        self.keys = collections.OrderedDict()

        self.initKeyBindings(key_values, key_translator)

        # Icon and button
        self.icon = None
        self._inactive_icon = None
        self._active_icon = None
        self.init_icon()
    # end Constructor
    
    def init_icon(self):
        """Create the icon."""
        try:
            self._active_icon = QtGui.QIcon(self.KEYBOARD_ACTIVE_IMG)
        except:
            self._active_icon = QtGui.QIcon()
            
        try:
            self._inactive_icon = QtGui.QIcon(self.KEYBOARD_INACTIVE_IMG)
        except:
            self._inactive_icon = QtGui.QIcon()
 
        self.icon = QtWidgets.QPushButton(self._inactive_icon, "", None)
        self.icon.is_gamepad = self
        self.icon.clicked.connect(self.toggle_active)
        self.icon.setToolTip(self.__class__.__name__ + " using the keyboard.")
    # end init_icon
    
    def get_init(self):
        """Return if the keybinding was initialized. (More of a compatibility method for the
        controllers).
        """
        return True
    
    def is_active(self):
        """Return if the key trigger is active."""
        return self._is_active
    
    def set_active(self, value=True):
        """Set if the keyboard should be active or not."""
        if value:
            self._is_active = True
            if threading.current_thread() == threading.main_thread():
                self.icon.setIcon(self._active_icon)
        else:
            self._is_active = False
            if threading.current_thread() == threading.main_thread():
                self.icon.setIcon(self._inactive_icon)
        
    def toggle_active(self):
        """Toggle the active state on or off."""
        self.set_active(not self.is_active())

    def initKeyBindings(self, key_values=None, key_translator=None):
        """Initialize the key bindings."""
        if key_values is None:
            key_values = self.DEFAULT_KEYS
        if key_translator is None:
            key_translator = self.DEFAULT_TRANSLATOR

        # Deep copy does not work for QtCore.Qt.Key_...
        for key in key_values:
            self.keys[key] = key_values[key]

        # Deep copy does not work for QtCore.Qt.Key_...    
        for key in key_translator:
            self.key_translator[key] = key_translator[key]
    # end initKeyBindings

    def set_key_value(self, key, value, send_key=True):
        """Set a specific key's value."""
        self.keys[key] = value
        if send_key: # Not repeat or allow sending repeat
            self.keys_pressed.emit(copy.deepcopy(self.keys))
    # end set_key_value

    def on_key_press(self, key, wasautorepeat=False):
        """When one of the set keys is pressed modify the parameters and emit the signal."""
        if not self.is_active():
            return

        # Modify the value from the second part of the key
        modify = 1
        send_autorepeat = True
        if isinstance(key, (tuple, list)):
            send_autorepeat = (len(key) <= 3 or key[3])
            # Check for a modify value and do not modify value if this was an autorepeat
            if len(key) > 1:
                modify = key[1]
            key = key[0]

        # Set the key value
        value = self.keys[key]
        if not wasautorepeat:
            value += modify
        self.set_key_value(key, value, not wasautorepeat or send_autorepeat)
    # end on_key_press

    def on_key_release(self, key, wasautorepeat=False):
        """When one of the set keys is released modify the parameters and emit the signal."""
        if not self.is_active():
            return

        # Check for send on release
        modify = 1
        send_release = False
        if isinstance(key, (tuple, list)):
            send_release = (len(key) > 2 and key[2])
            # Check for a modify value and do not modify value if this was an autorepeat
            if len(key) > 1:
                modify = key[1]
            key = key[0]

        # Set the key value
        value = self.keys[key]
        if not wasautorepeat:
            value -= modify
        self.set_key_value(key, value, not wasautorepeat and send_release)
    # end on_key_release

    def keyPressEvent(self, event):
        """Press a key."""
        key = self.key_translator.get(event.key(), None)
        if key is not None:
            self.on_key_press(key, event.isAutoRepeat())
            event.accept()
            return True
        else:
            return super().keyPressEvent(event)
    # end keyPressEvent

    def keyReleaseEvent(self, event):
        """Release a key."""
        key = self.key_translator.get(event.key(), None)
        if key is not None:
            self.on_key_release(key, event.isAutoRepeat())
            event.accept()
        return super().keyReleaseEvent(event)
    # end keyReleaseEvent

    def eventFilter(self, receiver, event):
        """Event filter for key press and release setting the key values and sending the signal."""
        etype = event.type()
        if etype == QtCore.QEvent.KeyPress:
            # Set the key value
            key = self.key_translator.get(event.key(), None)
            if key is not None:
                self.on_key_press(key, event.isAutoRepeat())
                return True

        elif etype == QtCore.QEvent.KeyRelease:
            # Release the key value
            key = self.key_translator.get(event.key(), None)
            if key is not None:
                self.on_key_release(key, event.isAutoRepeat())
                return True

        elif etype == QtCore.QEvent.FocusIn:
            receiver.grabKeyboard()
            return True
        elif etype == QtCore.QEvent.FocusOut:
            receiver.releaseKeyboard()
            return True

        return super().eventFilter(receiver, event)


class WASDController(BindKeys):
    DEFAULT_KEYS = collections.OrderedDict([("w", 0), ("a", 0), ("s", 0), ("d", 0)])
    DEFAULT_TRANSLATOR = {QtCore.Qt.Key_W: "w", QtCore.Qt.Key_A: "a", 
                          QtCore.Qt.Key_S: "s", QtCore.Qt.Key_D: "d"}


class ArrowController(BindKeys):
    DEFAULT_KEYS = collections.OrderedDict([("up", 0), ("right", 0), ("down", 0), ("left", 0), ("zoom", 0)])
    DEFAULT_TRANSLATOR = {QtCore.Qt.Key_Up: "up", QtCore.Qt.Key_Down: "down",
                          QtCore.Qt.Key_Left: "left", QtCore.Qt.Key_Right: "right"}


class GimbalController(BindKeys):
    DEFAULT_KEYS = collections.OrderedDict([("pan", 0), ("tilt", 0), ("zoom", 0), ("command", 0)])
    DEFAULT_TRANSLATOR = {QtCore.Qt.Key_Up: ("tilt", 0.2, True), QtCore.Qt.Key_Down: ("tilt", -0.2, True),
                          QtCore.Qt.Key_Left: ("pan", -0.2, True), QtCore.Qt.Key_Right: ("pan", 0.2, True),
                          QtCore.Qt.Key_Minus: ("zoom", -1, True),
                          QtCore.Qt.Key_Equal: ("zoom", 1, True),
                          QtCore.Qt.Key_1: ("command", 1, True),  # Toggle IR
                          QtCore.Qt.Key_2: ("command", 2, True),  # Toggle Laser
                          }



from .stash import Stash


__all__ = ['KeyTypes', 'HatValues', 'Key', 'Joystick']


class KeyTypes:
    AXIS = "Axis"
    BUTTON = "Button"
    HAT = "Hat"
    BALL = "Ball"
    ALL_KEYTYPES = ','.join((AXIS, BUTTON, HAT, BALL))

    @classmethod
    def has_keytype(cls, keytype, key_types):
        try:
            if keytype == cls.ALL_KEYTYPES or str(keytype) in key_types:
                return True
        except (TypeError, ValueError, Exception):
            pass
        return False


class HatValues:
    HAT_CENTERED = 0
    HAT_UP = 1
    HAT_RIGHT = 2
    HAT_DOWN = 4
    HAT_LEFT = 8
    HAT_RIGHTUP = HAT_UPRIGHT = 3
    HAT_RIGHTDOWN = HAT_DOWNRIGHT = 6
    HAT_LEFTUP = HAT_UPLEFT = 9
    HAT_LEFTDOWN = HAT_DOWNLEFT = 12

    ALL_HAT_VALUES = (HAT_CENTERED | HAT_UP | HAT_RIGHT | HAT_DOWN | HAT_LEFT |
                      HAT_UPRIGHT | HAT_DOWNRIGHT | HAT_UPLEFT | HAT_DOWNLEFT)

    HAT_CONVERTER = {
        HAT_CENTERED: 'Centered', HAT_UP: 'Up', HAT_RIGHT: 'Right', HAT_DOWN: 'Down', HAT_LEFT: 'Left',
        HAT_UPRIGHT: 'Up Right', HAT_DOWNRIGHT: 'Down Right', HAT_UPLEFT: 'Up Left', HAT_DOWNLEFT: 'Down Left',
        }

    NAME_CONVERTER = {name: value for value, name in HAT_CONVERTER.items()}

    @classmethod
    def convert_to_hat_name(cls, hat_value):
        """Return the given hat_value as a string name"""
        return cls.HAT_CONVERTER.get(hat_value, str(hat_value))

    @classmethod
    def convert_to_hat_value(cls, hat_name):
        """Return the given hat_name as an integer value. If -1 is returned it is an invalid value."""
        try:
            value = int(hat_name)
        except (TypeError, ValueError, Exception):
            value = -1
        return cls.NAME_CONVERTER.get(hat_name, value)


class Key(object):
    # Key Types
    KeyTypes = KeyTypes
    AXIS = KeyTypes.AXIS
    BUTTON = KeyTypes.BUTTON
    HAT = KeyTypes.HAT
    BALL = KeyTypes.BALL
    ALL_KEYTYPES = KeyTypes.ALL_KEYTYPES
    has_keytype = staticmethod(KeyTypes.has_keytype)

    # HAT Values (Guessing they are more bit flags than enums.)
    HatValues = HatValues
    HAT_CENTERED = HatValues.HAT_CENTERED
    HAT_UP = HatValues.HAT_UP
    HAT_RIGHT = HatValues.HAT_RIGHT
    HAT_DOWN = HatValues.HAT_DOWN
    HAT_LEFT = HatValues.HAT_LEFT
    HAT_RIGHTUP = HAT_UPRIGHT = HatValues.HAT_UPRIGHT
    HAT_RIGHTDOWN = HAT_DOWNRIGHT = HatValues.HAT_DOWNRIGHT
    HAT_LEFTUP = HAT_UPLEFT = HatValues.HAT_UPLEFT
    HAT_LEFTDOWN = HAT_DOWNLEFT = HatValues.HAT_DOWNLEFT
    ALL_HAT_VALUES = HatValues.ALL_HAT_VALUES
    convert_to_hat_name = staticmethod(HatValues.convert_to_hat_name)
    convert_to_hat_value = staticmethod(HatValues.convert_to_hat_value)

    def __init__(self, keytype, number, value=None, joystick=None, is_repeat=False, override=False):
        self.keytype = keytype
        self.number = number
        self.raw_value = None
        self.joystick = joystick
        self.is_repeat = is_repeat
        self.override = override

        self.set_value(value)

    def get_hat_name(self):
        """Return the value as a HAT name."""
        if self.keytype != self.HAT:
            raise TypeError('The Key must be a HAT keytype in order to get the hat name.')
        return self.convert_to_hat_name(self.raw_value)

    def get_proper_value(self):
        """Return the value between -1 and 1. Hat values act like buttons and will be 1 or 0.
        Use get_hat_name to check the keytype.
        """
        if self.raw_value is None:
            return 0
        elif self.raw_value > 1:
            return 1
        return self.raw_value

    def get_value(self):
        """Return the value of the key"""
        if self.raw_value is None:
            return 0
        return self.raw_value

    def set_value(self, value):
        """Set the value of the key"""
        self.raw_value = value

    value = property(get_value, set_value)

    def update_value(self, joystick=None):
        """Set this key's value from the set or given joystick's associated key value."""
        if joystick is None:
            joystick = self.joystick
        try:
            v = joystick.get_key(self).get_value()
            self.value = v
        except:
            pass

    def copy(self):
        """Create a copy of the key."""
        return self.__class__(self.keytype, self.number, self.value, self.joystick,
                              is_repeat=False, override=self.override)

    @classmethod
    def to_keyname(cls, key):
        """Return this key as a string keyname.

          * Format is "{minus}{keytype} {number}".
          * Hat format is "{keytype} {number} {hat_name}"

        Examples
            * "Axis 0" - For Axis 0 with a positive or 0 value.
            * "-Axis 1" - For an Axis Key that has a negative value and needs to be inverted.
            * "Button 0" - Buttons wont have negative values
            * "Hat 0 [Left Up]" - Hat values also give the key value as a hat name.
        """
        prefix = ''
        if key.value and key.value < 0:
            prefix = '-'

        if key.keytype == cls.HAT:
            return '{}{} {} [{}]'.format(prefix, key.keytype, key.number, key.get_hat_name())
        else:
            return '{}{} {}'.format(prefix, key.keytype, key.number)

    @classmethod
    def from_keyname(cls, keyname, joystick=None):
        """Return a new key from the given keyname."""
        # Remove any joystick name attached
        keyname = str(keyname)
        if ':' in keyname:
            keyname = keyname.split(':', 1)[-1].strip()

        # Split the keyname
        keytype, number = keyname.split(' ', 1)

        # Check if the keyname starts with a negative.
        value = None
        if keytype.startswith('-'):
            value = -1
            keytype = keytype[1:].strip()

        # Check if the number has '['
        if '[' in number:
            number, hat_name = number.split('[', 1)
            number = number.strip()
            value = int(cls.convert_to_hat_value(hat_name.replace(']', '').strip()))
        number = int(number)

        return Key(keytype, number, value, joystick=joystick)

    @property
    def keyname(self):
        return self.to_keyname(self)

    @keyname.setter
    def keyname(self, keyname):
        new_key = self.from_keyname(keyname)
        self.keytype = new_key.keytype
        self.number = new_key.number
        if self.value:
            self.value = new_key.value

    def __str__(self):
        return self.to_keyname(self)

    def __repr__(self):
        if self.joystick:
            return '<{module}.{name} {joystick}: {keyname} at {id}>'.format(
                    module=self.__module__, name=self.__class__.__name__, id=id(self),
                    joystick=self.joystick, keyname=self.keyname)
        else:
            return '<{module}.{name} {keyname} at {id}>'.format(
                    module=self.__module__, name=self.__class__.__name__, id=id(self),
                    joystick=self.joystick, keyname=self.keyname)

    def __hash__(self):
        return hash('{} {}'.format(self.keytype, self.number))

    def __eq__(self, other):
        try:
            if other.keytype == self.keytype and other.number == self.number:
                # Check if joysticks match if they are not None
                if other.joystick is not None and self.joystick is not None:
                    return other.joystick == self.joystick
                return True
            return False
        except:
            pass
        try:
            return str(self) == str(other)
        except:
            return False


class Joystick(object):
    @classmethod
    def get_joysticks(cls):
        """Return a list of available joysticks."""
        # return []
        raise NotImplementedError

    def __init__(self, *args, **kwargs):
        super().__init__()

        # Optional predefined variables (use with __new__)
        self.joystick = getattr(self, 'joystick', None)  # Internal joystick object

        self.identifier = getattr(self, 'identifier', -1)
        self.name = getattr(self, 'name', '')
        self.numaxes = getattr(self, 'numaxes', -1)
        self.numbuttons = getattr(self, 'numbuttons', -1)
        self.numhats = getattr(self, 'numhats', -1)
        self.numballs = getattr(self, 'numballs', -1)

        self.axis = getattr(self, 'axis', Stash())
        self.button = getattr(self, 'button', Stash())
        self.hat = getattr(self, 'hat', Stash())
        self.ball = getattr(self, 'ball', Stash())
        self.keys = getattr(self, 'keys', Stash(self.axis + self.button + self.hat + self.ball))

        self.deadband = getattr(self, 'deadband', 0.2)

        self.init_keys()

    def init_keys(self):
        """Initialize the keys."""
        self.axis = Stash(Key(Key.AXIS, i, None, self) for i in range(self.get_numaxes()))
        self.button = Stash(Key(Key.BUTTON, i, None, self) for i in range(self.get_numbuttons()))
        self.hat = Stash(Key(Key.HAT, i, None, self) for i in range(self.get_numhats()))
        self.ball = Stash(Key(Key.BALL, i, None, self) for i in range(self.get_numballs()))
        self.keys = Stash(self.axis + self.button + self.hat + self.ball)

    def is_available(self):
        """Return if this joystick is still active and available."""
        raise NotImplementedError

    def close(self):
        """Close the joystick."""
        raise NotImplementedError

    def get_key(self, key):
        """Return the key for the given key."""
        key_attr = getattr(self, str(key.keytype).lower())  # self.axis, self.button, self.hat, or self.ball
        return key_attr[key.number]

    def update_key(self, key):
        """Update the value for a given key."""
        self.get_key(key).set_value(key.value)

    def get_id(self):
        """Return the joystick id."""
        return self.identifier

    def get_name(self):
        """Return the name of the joystick."""
        return self.name

    def get_numaxes(self):
        """Return the number of axes."""
        return self.numaxes

    def get_axis(self, number):
        """Return the current value for the given axes."""
        return self.axis[number].get_value()

    def get_numbuttons(self):
        """Return the number of buttons."""
        return self.numbuttons

    def get_button(self, number):
        """Return the value for the given button number."""
        return self.button[number].get_value()

    def get_numhats(self):
        """Return the number of hats."""
        return self.numhats

    def get_hat(self, number):
        """Return the (hat [0], hat [1]) value for the given hat number."""
        return self.hat[number].get_value()

    def get_numballs(self):
        """Return the number of track balls."""
        return self.numballs

    def get_ball(self, number):
        """Return the current value for the given axes."""
        return self.ball[number].get_value()

    def get_deadband(self):
        """Return the deadband for this joystick axis."""
        return self.deadband

    def set_deadband(self, value):
        """Return the deadband for this joystick axis."""
        self.deadband = value

    def __eq__(self, other):
        name, ident, joystick = self.get_name(), self.get_id(), self.joystick
        try:
            return name == other.get_name() or ident == other or (joystick == other.joystick and joystick is not None)
        except:
            pass
        try:
            return name == other or ident == other or (joystick == other and joystick is not None)
        except:
            pass
        return False

    def __int__(self):
        return self.get_id()

    def __str__(self):
        return self.get_name()

    def __hash__(self):
        return hash('{} {}'.format(self.identifier, self.name))

    def __getstate__(self):
        return {
            'joystick': None,
            'identifier': self.identifier,
            'name': self.name,
            'numaxes': self.numaxes,
            'numbuttons': self.numbuttons,
            'numhats': self.numhats,
            'numballs': self.numballs,

            'axis': self.axis,
            'button': self.button,
            'hat': self.hat,
            'ball': self.ball,
            'keys': self.keys,

            'deadband': self.deadband
            }

    def __setstate__(self, state):
        for k, v in state.items():
            setattr(self, k, v)
